package jadex.tools.jcc;

import jadex.base.Starter;
import jadex.base.gui.CMSUpdateHandler;
import jadex.base.gui.SwingDefaultResultListener;
import jadex.base.gui.SwingDelegationResultListener;
import jadex.base.gui.plugin.SJCC;
import jadex.bridge.IExternalAccess;
import jadex.bridge.service.RequiredServiceInfo;
import jadex.bridge.service.SServiceProvider;
import jadex.bridge.service.library.ILibraryService;
import jadex.commons.Properties;
import jadex.commons.Property;
import jadex.commons.future.Future;
import jadex.commons.future.IFuture;
import jadex.commons.gui.SGUI;
import jadex.xml.PropertiesXMLHelper;

import java.awt.Dimension;
import java.awt.Toolkit;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Map;

import javax.swing.JOptionPane;
import javax.swing.SwingUtilities;

/**
 *  The global control center.
 */
public class ControlCenter
{
	// -------- constants --------

	/** The filename extension for GUI settings. */
	public static final String	SETTINGS_EXTENSION	= ".settings.xml";

	/** Auto-shutdown on exit. */
	public static final String	JCC_EXIT_SHUTDOWN	= "shutdown";

	/** No shutdown on exit. */
	public static final String	JCC_EXIT_KEEP	= "keep";

	/** Ask for shutdown on exit. */
	public static final String	JCC_EXIT_ASK	= "ask";

	// -------- attributes --------

	/** The jcc component. */
	protected IExternalAccess jccaccess;
	
	/** The plugin classes. */
	// Todo: load classes only once!?
	protected String[]	plugin_classes;
	
	/** The platform control centers (cid -> pcc). */
	protected Map	pccs;

	/** The currently displayed platform control center. */
	protected PlatformControlCenter	pcc;

	/** The control center window. */
	protected ControlCenterWindow	window;
	
	/** Flag indicating if exit was initiated. */
	protected boolean	killed;

	/** Shutdown action on exit (ask (default), keep, shutdown). */
	protected String	jccexit;

	/** The CMS update handler shared by all tools. */
	protected CMSUpdateHandler	cmshandler;
	
	/** The save on exit flag. */
	protected boolean	saveonexit;

	//-------- constructors --------

	/**
	 *  Create a control center.
	 */
	public IFuture	init(IExternalAccess jccaccess, final String[] plugin_classes, boolean saveonexit)
	{
		final Future	ret	= new Future();
		
		this.jccaccess = jccaccess;
		this.plugin_classes	= plugin_classes;
		this.pccs	= new HashMap();
		this.saveonexit	= saveonexit;
		this.window = new ControlCenterWindow(this);
		
		// Default platform control center for local platform.
		final Future	inited	= new Future();
		this.pcc	= new PlatformControlCenter();
		
		SJCC.getRootAccess(jccaccess).addResultListener(new SwingDelegationResultListener(inited)
		{
			public void customResultAvailable(Object result)
			{
				IExternalAccess	platformaccess	= (IExternalAccess)result;
				pccs.put(platformaccess.getComponentIdentifier(), pcc);
				pcc.init(platformaccess, ControlCenter.this, plugin_classes)
					.addResultListener(new SwingDelegationResultListener(inited));
			}
		});
		
		inited.addResultListener(new SwingDelegationResultListener(ret)
		{
			public void customResultAvailable(Object result)
			{
				// Load settings and open window.
				loadSettings().addResultListener(new SwingDelegationResultListener(ret)
				{
					public void customResultAvailable(Object result)
					{
						// Add PCC to window.
						window.showPlatformPanel(pcc);
						window.setVisible(true);
						ret.setResult(null);
					}
				});
			}
		});
		
		return ret;
	}

	//-------- methods called by control center window --------

	/**
	 *  Load the settings.
	 */
	public IFuture	loadSettings()
	{
		// Only load GUI settings as platform settings are loaded on platform startup and every tool init as required.
		return loadSettings(new File(jccaccess.getComponentIdentifier().getLocalName() + SETTINGS_EXTENSION));
	}
	
	/**
	 *  Load the settings.
	 */
	public IFuture	loadSettings(final File file)
	{
		final Future	ret	= new Future();
		
		// Read project properties
		SServiceProvider.getService(jccaccess.getServiceProvider(), ILibraryService.class, RequiredServiceInfo.SCOPE_PLATFORM)
			.addResultListener(new SwingDelegationResultListener(ret)
		{
			public void customResultAvailable(Object result)
			{
				Properties	props;
				try
				{
					ClassLoader cl = ((ILibraryService)result).getClassLoader();
					FileInputStream fis = new FileInputStream(file);
					props	= (Properties)PropertiesXMLHelper.getPropertyReader().read(fis, cl, null);
					fis.close();
					
					Properties windowprops = props.getSubproperty("window");
					if(windowprops != null)
					{
						int w = windowprops.getIntProperty("width");
						int h = windowprops.getIntProperty("height");
						int x = windowprops.getIntProperty("x");
						int y = windowprops.getIntProperty("y");
						window.setBounds(x, y, w, h);
			
						window.setVisible(true); // otherwise it will not be extended (jdk5)
						int es = windowprops.getIntProperty("extendedState");
						window.setExtendedState(es);
			
						jccexit = windowprops.getStringProperty("jccexit");
						
						// Do not override saveonexit agent argument to true by loaded properties.
						saveonexit = saveonexit && windowprops.getBooleanProperty("saveonexit");
					}
				}
				catch(Exception e)
				{
					// Use default values when settings cannot be loaded.
					props	= new Properties();
					
					Dimension	screendim = Toolkit.getDefaultToolkit().getScreenSize();
					// 60% of screen but not smaller than 800x650 but not exceeding screen.
					Dimension	windim	= new Dimension(
						Math.min(Math.max((int)(screendim.width * 0.6), 800), screendim.width),
						Math.min(Math.max((int)(screendim.height * 0.6), 650), screendim.height));
					window.setSize(windim);
					window.setLocation(SGUI.calculateMiddlePosition(window));
				}
				
				pcc.setProperties(props).addResultListener(new SwingDelegationResultListener(ret));
			}
		});
		
		return ret;
	}

	/**
	 * Save settings of JCC and all plugins in current project.
	 */
	public IFuture	saveSettings()
	{
		final Future	ret	= new Future();
//		System.out.println("Saving JCC settings");
		// Save settings of GUI and currently selected platform (todo: all platforms?)
		saveSettings(new File(jccaccess.getComponentIdentifier().getLocalName() + SETTINGS_EXTENSION))
			.addResultListener(new SwingDelegationResultListener(ret)
		{
			public void customResultAvailable(Object result)
			{
//				System.out.println("Saving platform settings");
				pcc.savePlatformProperties().addResultListener(new SwingDelegationResultListener(ret));
			}
		});
		
		return ret;
	}

	/**
	 * Save settings of JCC and all plugins in current project.
	 */
	public IFuture	saveSettings(final File file)
	{
		final Future	ret	= new Future();
//		System.out.println("Fetching JCC properties.");
		
		// Get properties of latest platform panel.
		pcc.getProperties().addResultListener(new SwingDelegationResultListener(ret)
		{
			public void customResultAvailable(Object result)
			{
//				System.out.println("Fetched JCC properties.");
				final Properties	props	= (Properties)result;
			
				// Store window appearance
				Properties windowprops = new Properties();
				windowprops.addProperty(new Property("width", Integer.toString(window.getWidth())));
				windowprops.addProperty(new Property("height", Integer.toString(window.getHeight())));
				windowprops.addProperty(new Property("x", Integer.toString(window.getX())));
				windowprops.addProperty(new Property("y", Integer.toString(window.getY())));
				// getExtendedState() deadlocks when shutdown hook is triggered.
				windowprops.addProperty(new Property("extendedState", Integer.toString(window.getCachedState())));
				windowprops.addProperty(new Property("jccexit", jccexit != null? jccexit : JCC_EXIT_ASK));
				windowprops.addProperty(new Property("saveonexit", Boolean.toString(saveonexit)));
				props.removeSubproperties("window");
				props.addSubproperties("window", windowprops);
				
//				System.out.println("Fetching library service.");
				// Save properties to file.
				SServiceProvider.getService(jccaccess.getServiceProvider(), ILibraryService.class, RequiredServiceInfo.SCOPE_PLATFORM)
					.addResultListener(new SwingDelegationResultListener(ret)
				{
					public void customResultAvailable(Object result)
					{
						try
						{
//							System.out.println("Writing properties.");
							FileOutputStream os = new FileOutputStream(file);
							PropertiesXMLHelper.getPropertyWriter().write(props, os, ((ILibraryService)result).getClassLoader(), null);
							os.close();
							window.getStatusBar().setText("Settings saved successfully: "+ file.getAbsolutePath());
							ret.setResult(null);
						}
						catch(Exception e)
						{
//							e.printStackTrace();
							throw new RuntimeException(e);
						}
					}
				});
			}
		});
		
		return ret;
	}

	/**
	 * Informs the window if it should dispose its resources.
	 * @return true if the agent has been killed.
	 */
	public void	exit()
	{
		assert SwingUtilities.isEventDispatchThread() ||  Starter.isShutdown();
		
		if(!killed)
		{
			int choice;
			RememberOptionMessage msg = null;
			if(jccexit==null || jccexit.equals(JCC_EXIT_ASK))
			{
				msg = new RememberOptionMessage("You requested to close the JCC GUI.\n "
					+ "Do you also want to shutdown the local platform?\n");
				choice = JOptionPane.showConfirmDialog(window, msg, "Exit Question",
					JOptionPane.YES_NO_CANCEL_OPTION, JOptionPane.QUESTION_MESSAGE);
			}
			else if(jccexit.equals(JCC_EXIT_KEEP))
			{
				choice = JOptionPane.NO_OPTION;
			}
			else // if(jccexit.equals(JCC_EXIT_SHUTDOWN))
			{
				choice = JOptionPane.YES_OPTION;
			}
	
			if(JOptionPane.YES_OPTION == choice)
			{
				// Save settings if wanted
				if(msg!=null && msg.isRemember())
					jccexit = JCC_EXIT_SHUTDOWN;
	
				SJCC.killPlattform(jccaccess, window);
			}
			else if(JOptionPane.NO_OPTION == choice)
			{
				// Save settings if wanted
				if(msg != null && msg.isRemember())
					jccexit = JCC_EXIT_KEEP;
	
				jccaccess.killComponent();
			}
			// else CANCEL
		}
	}
	
	/**
	 *  Do any required cleanup on exit.
	 */
	public IFuture	shutdown()
	{
//		System.out.println("Control Center shutdown A.");
		final Future	ret	= new Future();
		
		Runnable	runnable	= new Runnable()
		{
			public void run()
			{
//				System.out.println("Control Center shutdown B.");
				assert !killed;
				killed = true;
				
				IFuture	saved;
				if(saveonexit)
					saved	= saveSettings();
				else
					saved	= IFuture.DONE;
				
				saved.addResultListener(new SwingDelegationResultListener(ret)
				{
					public void customResultAvailable(Object result)
					{
//						System.out.println("JCC settings saved.");
						
						// Todo: pcc dispose with future?
						for(Iterator it=pccs.keySet().iterator(); it.hasNext(); )
						{
							((PlatformControlCenter)pccs.get(it.next())).dispose();
						}
						
						IFuture	handlerdisposed;
						if(cmshandler!=null)
							handlerdisposed	= cmshandler.dispose();
						else
							handlerdisposed	= IFuture.DONE;

						handlerdisposed.addResultListener(new SwingDelegationResultListener(ret)
						{
							public void customResultAvailable(Object result)
							{
//								System.out.println("CMS handlers disposed.");
								if(!Starter.isShutdown())
								{
									window.setVisible(false);
									window.dispose();
								}
								ret.setResult(null);
							}
						});
					}
				});
			}
		};
		
		// Hack!!! When triggered from shutdown hook, swing might be terminated
		// and invokeLater has no effect (grrr).
		if(Starter.isShutdown())
		{
			runnable.run();
		}
		else
		{
			SwingUtilities.invokeLater(runnable);
		}
			
		
		return ret;
	}
	
	/**
	 *  Get the save on exit flag.
	 */
	public boolean isSaveOnExit()
	{
		return saveonexit;
	}
	
	/**
	 *  Set the save on exit flag.
	 */
	public void setSaveOnExit(boolean saveonexit)
	{
		this.saveonexit	= saveonexit;
	}
	
	//-------- methods used by platform control centers --------
	
	/**
	 *  Get the cms update handler shared by all tools.
	 */
	public CMSUpdateHandler getCMSHandler()
	{
		assert SwingUtilities.isEventDispatchThread() ||  Starter.isShutdown();
		
		if(cmshandler==null)
		{
			cmshandler	= new CMSUpdateHandler(jccaccess);
		}
		return cmshandler;
	}
	
	/**
	 *  Get the control center window.
	 */
	public ControlCenterWindow	getWindow()
	{
		return window;
	}
	
	/**
	 *  Get the JCC access.
	 */
	public IExternalAccess	getJCCAccess()
	{
		return jccaccess;
	}
	
	/**
	 *  Add a new platform control center
	 *  or switch to tab if already exists.
	 */
	public void	showPlatform(IExternalAccess platformaccess)
	{
		final Future	newpcc	= new Future();
		if(pccs.containsKey(platformaccess.getComponentIdentifier()))
		{
			newpcc.setResult(pccs.get(platformaccess.getComponentIdentifier()));
		}
		else
		{
			final PlatformControlCenter	tmp	= new PlatformControlCenter();
			pccs.put(platformaccess.getComponentIdentifier(), tmp);
			tmp.init(platformaccess, this, plugin_classes)
				.addResultListener(new SwingDelegationResultListener(newpcc)
			{
				public void customResultAvailable(Object result)
				{
					newpcc.setResult(tmp);
				}
			});
		}
		
		newpcc.addResultListener(new SwingDefaultResultListener(window)
		{
			public void customResultAvailable(Object result)
			{
				final PlatformControlCenter	pcc2	= (PlatformControlCenter)result;
				
				// Transfer settings from pcc to pcc2
				pcc.getProperties().addResultListener(new SwingDefaultResultListener(window)
				{
					public void customResultAvailable(Object result)
					{
						pcc2.setProperties((Properties)result).addResultListener(new SwingDefaultResultListener(window)
						{
							public void customResultAvailable(Object result)
							{
								pcc	= pcc2;
								window.showPlatformPanel(pcc);
							}
						});
					}
				});
				
			}
		});
	}

	/**
	 *  Close a platform control center.
	 */
	public void closePlatform(final PlatformControlCenter pcc)
	{
		pccs.remove(pcc.getPlatformAccess().getComponentIdentifier());
		window.getStatusBar().setText("Saving platform settings for: "+pcc.getPlatformAccess().getComponentIdentifier().getPlatformName());
		window.closePlatformPanel(pcc);
		
		IFuture	saved	= isSaveOnExit() ? pcc.savePlatformProperties() : IFuture.DONE;
		saved.addResultListener(new SwingDefaultResultListener()
		{
			public void customResultAvailable(Object result)
			{
				window.getStatusBar().setText("Saved platform settings for: "+pcc.getPlatformAccess().getComponentIdentifier().getPlatformName());
				pcc.dispose();
			}
			public void customExceptionOccurred(Exception exception)
			{
				// Continue anyways.
				window.getStatusBar().setText("Could not save platform settings: "+exception);
				pcc.dispose();
			}
		});
	}
	
	//-------- used for test case --------
	
	/**
	 *  Get the current platform control center.
	 */
	public PlatformControlCenter	getPCC()
	{
		return pcc;
	}
}
